//
//	Written by Dmitry Friesen, 20/03/2003
//	e-mail: dfriesen1946@mail.ru
//

//------------------------------------------------------------------------
//	Borland C++ Builder RTTI support for IDA
//

#include "IDA_Misc.idc"
#include "BCCtypes.idc"

// -----------------------------------------
//	make '@$xp$%d%s' names
static MakeDPDNames(ea, name)
{
	auto fea;

	name = form("@$xp$%d%s", strlen(name), name);
	if ( (fea = LocByName(name)) == BADADDR || fea == Dword(ea) )
		MyMakeName(Dword(ea), name);
	else {
		Message(form(" ---> ERROR: Make name '%s' at 0x%08lX exist at 0x%08lX\n", name, ea, fea));
	}
}

//
#define offs_tpMask		0x00000004L
#define offs_tpName		0x00000006L
#define offs_tpBase		0x00000008L
#define offs_tpcFlags	0x0000000CL
#define offs_Elements	0x0000000CL
#define offs_Size		0x00000010L
#define offs_DtorAddr	0x00000028L
#define offs_FullSize	0x0000002EL

//	tpMask flags
#define TM_IS_STRUCT    0x0001
#define TM_IS_CLASS     0x0002
#define TM_IS_UNK04     0x0004
#define TM_IS_UNK08     0x0008
#define TM_IS_PTR       0x0010
#define TM_IS_REF       0x0020
#define TM_IS_VOIDPTR   0x0040
#define TM_LOCALTYPE    0x0080
#define TM_IS_CONST     0x0100
#define TM_IS_VOLATILE  0x0200
#define TM_IS_ARRAY     0x0400

//	Class flags tpcFlags follow:
#define CF_HAS_CTOR     0x00000001
#define CF_HAS_DTOR     0x00000002
#define CF_HAS_BASES    0x00000004
#define CF_HAS_VBASES   0x00000008
#define CF_HAS_VTABPTR  0x00000010
#define CF_HAS_VIRTDT   0x00000020
#define CF_HAS_RTTI     0x00000040
#define CF_DELPHICLASS  0x00000080
#define CF_HAS_FARVPTR  0x00001000
#define CF_HAS_GUID     0x00002000  // The tpcGuid field is valid and filled in

//	kinds of name
#define tkSimple 1
#define tkPtr 2
#define tkRef 3
#define tkArr 4

// -----------------------------------------
//
static isTPDSC(ea)
{
	auto tpMask;

	if ( isCStr(ea + Word(ea + offs_tpName)) ) {
		tpMask = Word(ea + offs_tpMask);
		if ( tpMask == 0x0001 || tpMask == 0x0003 || tpMask == 0x0010 || tpMask == 0x0090 || tpMask == 0x00A0 || tpMask == 0x0400 )
			return 1;
	}
	return 0;
}

// -----------------------------------------
//	Get DelphiInterface name
static DelphiInter(ea, term)
{
	auto str, c;

	str = "";
	for ( c = Byte(ea); c != term; c = Byte(ea) ) {
		if ( c == ':' && Byte(ea + 1) == ':' ) {	// found namespace?
			c = "@";
			ea++;
		}
		str = str + c;
		ea++;
	}
	// TODO: '::' & '<>' in str
	return ("System@%DelphiInterface$t" + form("%d%s%%", strlen(str), str));
}

// -----------------------------------------
//	Get class name from tpid.
//	from: '_di_name'
//	  to: 'System@%DelphiInterface$tlname%'
//	from: namespace::name<str1,str2,...>
//	  to: namespace@%name$tlstr1lstr2...%
static GetClsName(ea)
{
	auto uea, str, c, name;

Message(form(" ---> GetClsName(0x%08lX)", ea));

	uea = ea;
	ea = ea + Word(ea + offs_tpName);

	if ( Byte(ea) == '_' && Byte(ea + 1) == 'd' && Byte(ea + 2) == 'i' && Byte(ea + 3) == '_' )
		return DelphiInter(ea + 4, 0);

	str = "";
	for ( c = Byte(ea); c != 0; c = Byte(ea) ) {
		if ( c == '<' ) {
			if ( strstr(str, "DelphiInterface") != -1 ) {
				str = DelphiInter(ea + 1, '>');
				break;
			}
Message(" return unknown mangled class name\n");
			return form("Unk_%lX", uea);
		}
		if ( c == ':' && Byte(ea + 1) == ':' ) {	// found namespace?
			c = "@";
			ea++;
		}
		str = str + c;
		ea++;
	}
Message(form(" return '%s'\n", str));
	return str;
}

// -----------------------------------------
//	make a Borland mangled name of typeinfo
static GetBTName(ea)
{
	auto tpMask, name;

	tpMask = Word(ea + offs_tpMask);
	if ( tpMask & TM_IS_PTR )
		return ("p" + GetBTName(Dword(ea + offs_tpBase)));
	else if ( tpMask & TM_IS_REF )
		return ("r" + GetBTName(Dword(ea + offs_tpBase)));
	else if ( tpMask & TM_IS_ARRAY )
		return (form("a%d$", Dword(ea + offs_Elements)) + GetBTName(Dword(ea + offs_tpBase)));
	else if ( tpMask & TM_IS_CLASS ) {
		name = GetClsName(ea);
		return form("%d%s", strlen(name), name);
	}
}

// -----------------------------------------
//	parse a list of base classes
static doBases(ea, comment)
{ 
	if ( Dword(ea) ) {
		MakeComm(ea, comment);
		while ( Dword(ea) ) {
			ea = MyMakeStruct(ea, "_tpcBase");
		}
		MakeComm(ea, "eol");
		ea = MyMakeDword(ea);
	} else {
		MakeComm(ea, comment + ": eol");
		ea = MyMakeDword(ea);
	}
	return ea;
}

// -----------------------------------------
//	parse a list of destructible members
static doMembers(ea, comment)
{ 
	if ( Dword(ea) ) {
		MakeComm(ea, comment);
		while ( Dword(ea) ) {
			ea = MyMakeStruct(ea, "_tpcMemb");
		}
		MakeComm(ea, "eol");
		ea = MyMakeDword(ea);
	} else {
		MakeComm(ea, comment + ": eol");
		ea = MyMakeDword(ea);
	}
	return ea;
}

// -----------------------------------------
//
static MakeTPDSC(ea)
{
	auto tpMask, tpcFlags, end, cea, pfn, off, name;

	tpMask = Word(ea + offs_tpMask);
	if ( tpMask & TM_IS_PTR ) {
		cea = MyMakeStruct(ea, "_tpidp");
		MakeCStr(cea, "");
		cea = ItemEnd(cea);
		name = GetBTName(ea);
		if ( name != "" )
			MyMakeName(ea, "@$xt$" + name);
	} else if ( tpMask & TM_IS_REF ) {
		cea = MyMakeStruct(ea, "_tpidp");
		MakeCStr(cea, "");
		cea = ItemEnd(cea);
		name = GetBTName(ea);
		if ( name != "" )
			MyMakeName(ea, "@$xt$" + name);
	} else if ( tpMask & TM_IS_ARRAY ) {
		cea = MyMakeStruct(ea, "_tpida");
		MakeCStr(cea, "");
		cea = ItemEnd(cea);
		name = GetBTName(ea);
		if ( name != "" )
			MyMakeName(ea, "@$xt$" + name);
	} else if ( tpMask & TM_IS_CLASS ) {
		end = ea + Word(ea + offs_Size);
		tpcFlags = Dword(ea + offs_tpcFlags);
		if ( tpcFlags & CF_HAS_DTOR ) {
			if ( (pfn = Dword(ea + offs_DtorAddr)) ) {
				name = GetClsName(ea);
				if ( name != "" )
					MakeSub(pfn, form("@%s@$bdtr$q", name));	// parameter list unknown
			}
			cea = MyMakeStruct(ea, "_tpidc");
		} else {
			cea = MyMakeStruct(ea, "_tpidcs");
		}
		MakeCStr(cea, "");
		cea = ItemEnd(cea);
		name = GetBTName(ea);
		if ( name != "" )
			MyMakeName(ea, "@$xt$" + name);
		if ( cea < end )
			cea = MyMakeAlign(cea);
		if ( tpcFlags & CF_HAS_GUID ) {
			cea = MyMakeStruct(cea, "_GUID");
		}
		if ( (off = Word(ea + 16)) != 0 )		// tpcBaseList
			cea = doBases(ea + off, "Base classes");
		if ( (off = Word(ea + 18)) != 0 )		// tpcVbasList
			cea = doBases(ea + off, "Virtual base classes");
		if ( tpcFlags & CF_HAS_DTOR ) {
			if ( (off = Word(ea + 0x2E)) != 0 )	// tpcDtMembers
				cea = doMembers(ea + off, "Destructible members");
		}
	} else {
		Message(" ---> Unknown tpMask ");
		return BADADDR;
	}

	return cea;
}

// -----------------------------------------
//
static isInt(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 1 );
}

// -----------------------------------------
//	make Integer __DPdsc__
static MakeInt(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeStruct(ea, "_tcint");
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isChar(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 2 );
}

// -----------------------------------------
// make Char __DPdsc__
static MakeChar(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeStruct(ea, "_tcint");
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isEnum(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 3 );
}

// -----------------------------------------
//	make Enumeration __DPdsc__
static MakeEnum(ea, basename)
{
	auto start, classname, min, max, bea;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	min = Dword(ea + 1);
	max = Dword(ea + 5);
	bea = Dword(ea + 9);
	ea = MyMakeStruct(ea, "_tcenum");
	if ( bea == start ) {
		for ( ; min <= max; min++ ) {
			if ( !IsPasStr(ea) )
				return BADADDR;
			MakePasStr(ea);
			ea = ItemEnd(ea);
		}
	}
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isFloat(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 4 );
}

// -----------------------------------------
//	make Float __DPdsc__
static MyMakeFloat(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeByte(ea);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isString(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 5 );
}

// -----------------------------------------
//	make String __DPdsc__
static MyMakeString(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeComm(ea, "sizeof=");
	ea = MyMakeByte(ea);			// sizeof=
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isSet(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 6 );
}

// -----------------------------------------
//	make Set __DPdsc__
static MakeSet(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeStruct(ea, "_tcset");
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isClass(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 7 );
}

// -----------------------------------------
//	make Class __DPdsc__
static MakeClass(ea, basename)
{
	auto start, classname, name, nprops, pea, propName;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeStruct(ea, "_tcclass");
	basename = MakePasStr(ea);
	ea = ItemEnd(ea);
	name = basename + "@" + classname;
	nprops = Word(ea);
	MakeComm(ea, "PropCount");
	ea = MyMakeWord(ea);
	for ( ; nprops > 0; nprops-- ) {
		pea = ea + 4;
		ea = MyMakeStruct(ea, "TPropInfo");
		propName = MakePasStr(ea);
		ea = ItemEnd(ea);
		if ( IsContainPtr(pea) )	// GetProc
			MakeSub(Dword(pea), form("@%s@Get%s$qqrv", name, propName));
		pea = pea + 4;
		if ( IsContainPtr(pea) )	// SetProc
			MakeSub(Dword(pea), form("@%s@Set%s$qqrx", name, propName));
		pea = pea + 4;
		if ( IsContainPtr(pea) )	// StoredProc
			MakeSub(Dword(pea), form("@%s@Is%sStored$qqrv", name, propName));
	}
	MakeDPDNames(start, name);
	return ea;
}

// -----------------------------------------
//
static isMethod(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 8 ); 
}

// -----------------------------------------
//
static MakeMethod(ea, basename)
{
	auto start, classname, bRet, nparam;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	bRet = Byte(ea);
	if ( bRet )
		MakeComm(ea, "is Function");
	else
		MakeComm(ea, "is Procedure");
	ea = MyMakeByte(ea);
	nparam = Byte(ea);
	ea = MyMakeByte(ea);
	for ( ; nparam > 0; nparam-- ) {
		MakeComm(ea, "pfAddress");
		ea = MyMakeByte(ea);		// pfAddress
		MakeComm(ea, "ParamName");
		MakePasStr(ea);				// ParamName
		ea = ItemEnd(ea);
		MakeComm(ea, "ParamType");
		MakePasStr(ea);				// ParamType
		ea = ItemEnd(ea);
	}
	if ( bRet ) {
		MakeComm(ea, "ResultType");
		MakePasStr(ea);				// ResultType
		ea = ItemEnd(ea);
	}
	if ( classname == "" )
		classname = form("UnkMtd%lX", start);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);

	return ea;
}

// -----------------------------------------
//
static isLString(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0A ); 
}

// -----------------------------------------
//	make LString __DPdsc__
static MakeLString(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isWString(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0B ); 
}

// -----------------------------------------
//	make WString __DPdsc__
static MakeWString(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isVariant(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0C ); 
}

// -----------------------------------------
//	make Variant __DPdsc__
static MakeVariant(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isArray(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0D ); 
}

// -----------------------------------------
//	make Array __DPdsc__
static MyMakeArray(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeDword(ea);
	ea = MyMakeDword(ea);
	ea = MyMakeOff(ea);
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isRecord(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0E ); 
}

// -----------------------------------------
//	make Record __DPdsc__
static MakeRecord(ea, basename)
{
	auto start, classname, nbase;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeDword(ea);
	nbase = Dword(ea);
	ea = MyMakeDword(ea);
//	MakeComm(ea, "base record");
	for ( ; nbase > 0; nbase-- ) {
		ea = MyMakeStruct(ea, "_tcrec");		// type record field
	}
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isInterface(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x0F ); 
}

// -----------------------------------------
//	make Interface __DPdsc__
static MakeInterface(ea, basename)
{
	auto start, classname, ifguid;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeComm(ea, "base interface");
	ea = MyMakeOff(ea);		// base interface
	ifguid = Byte(ea);
	MakeComm(ea, "ifHasGuid");
	ea = MyMakeByte(ea);		// ifHasGuid
	if ( ifguid == 1 ) {
		ea = MyMakeStruct(ea, "_GUID");
	}
	basename = MakePasStr(ea);
	ea = ItemEnd(ea);
	ea = MyMakeWord(ea);		// PropCount
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isDynArray(ea)
{
	return ( Dword(ea) == ea + 4 && Byte(ea + 4) == 0x11 ); 
}

// -----------------------------------------
//	make DynArray __DPdsc__
static MakeDynArray(ea, basename)
{
	auto start, classname;

	start = ea;
	ea = MyMakeOff(ea);
	ea = MyMakeStruct(ea, "_tchd");
	classname = MakePasStr(ea);
	ea = ItemEnd(ea);
	MakeComm(ea, "sizeof(arr[0])");
	ea = MyMakeDword(ea);		// sizeof(arr[0])
	ea = MyMakeDword(ea);		// min
	ea = MyMakeDword(ea);		// max
	MakeDPDNames(start, (basename == ""?"System":basename) + "@" + classname);
	return ea;
}

// -----------------------------------------
//
static isBCCString(ea)
{
	auto n;

	if ( Dword(ea) == 0xFFFFFFFFL ) {
		ea = ea + 8;
		for ( n = Dword(ea - 4); n > 0 && Byte(ea) != 0; n-- ) {
			ea++;
		}
		return ( n == 0 && Byte(ea) == 0 );
	}
	return 0;
}

// -----------------------------------------
// 
static MakeBCCString(ea)
{
	ea = MyMakeStruct(ea, "_string");
	MakeCStr(ea, "");
	return ItemEnd(ea);
}

#define offs_VTBL	0x00000000L		// virtual table
#define offs_INTF	0x00000004L		// interface table
#define offs_XXX	0x00000008L		// ???
#define offs_PROPS	0x0000000CL		// properties table
#define offs_DTC	0x00000010L		// data type info
#define offs_ACLS	0x00000014L		// accelerators table
#define offs_MNUS	0x00000018L		// menus table
#define offs_MSG	0x0000001CL		// messages table
#define offs_CNAME	0x00000020L		// class name
#define offs_SIZE	0x00000024L		// size class data
#define offs_BRTC	0x00000028L		// base class
#define offs_PROCL	0x0000002CL		// base procedures
#define sizeof_RTC	0x0000004CL		// size of class table

// -----------------------------------------
//
static isRTDSC(ea)
{
	return ( Dword(ea) - ea == sizeof_RTC && IsPasStr(Dword(ea + offs_CNAME)) );
}

// -----------------------------------------
// make Interfaces
static MakeINTFs(ea)
{
	auto nintfs;

	nintfs = Dword(ea);
	ea = MyMakeDword(ea);
	for ( ; nintfs > 0; nintfs-- )
		ea = MyMakeStruct(ea, "_rtintf");

	return ea;
}

// -----------------------------------------
// make Properties
static MakePROPs(ea)
{
	auto nProps;

	ea = MyMakeWord(ea);
	ea = MyMakeDword(ea);
	nProps = Dword(ea);
	ea = MyMakeDword(ea);		// nprops
	for ( ; nProps > 0; nProps-- )
		ea = MyMakeStruct(ea, "_rtprop");

	return ea;
}

// -----------------------------------------
//
static MakeSBCs(ea)
{
	auto nsbcs;

	nsbcs = Word(ea);
	ea = MyMakeWord(ea);
	for ( ; nsbcs > 0; nsbcs-- )
		ea = MyMakeOff(ea);

	return ea;
}

// -----------------------------------------
//
static MakeACLs(ea)
{
	auto nacls, psbc;

	nacls = Word(ea);
	ea = MyMakeWord(ea);
	if ( IsContainPtr(ea) )
		psbc = MakeSBCs(Dword(ea));
	ea = MyMakeOff(ea);
	for ( ; nacls > 0; nacls-- ) {
		ea = MyMakeStruct(ea, "_acl");
		MakePasStr(ea);
		ea = ItemEnd(ea);
	}

	if ( ea < psbc )
		ea = psbc;

	return ea;
}

// -----------------------------------------
//
static MakeMNUs(ea)
{
	auto nmnus, pfn, pname;

	nmnus = Word(ea);
	ea = MyMakeWord(ea);
	for ( ; nmnus > 0; nmnus-- ) {
		ea = MyMakeStruct(ea, "_mnu");
		pname = MakePasStr(ea);
		ea = ItemEnd(ea);
	}

	return ea;
}

// -----------------------------------------
//
static MakeMSGs(ea)
{
	auto nmsgs, pfn;

	nmsgs = Word(ea);
	ea = MyMakeWord(ea);
	pfn = ea + nmsgs * 2;

	for ( ; nmsgs > 0; nmsgs-- ) {
		ea = MyMakeWord(ea);
		MakeSub(Dword(pfn), "");
		pfn = MyMakeOff(pfn);
	}

	return pfn;
}

// -----------------------------------------
//
static MakeRTDSC(ea)
{
	auto pclassname, classname, pvtable, pxxx, pintf, pdtclass, pprops, pacls, pmnus, pmsgs;
	auto min, cea, max;

	min = 0x7FFFFFFFL;
	max = 0;

	pvtable = MyMakeStruct(ea, "TRTCInfo");	// vtable address

	pclassname = Dword(ea + offs_CNAME);	// class name address
	classname = MakePasStr(pclassname);
	MyMakeName(pclassname, form("psz%s", classname));
	if ( (pclassname = ItemEnd(pclassname)) > max )
		max = pclassname;

	for ( cea = ea + 4; cea < pvtable; cea = cea + 4 ) {
		if ( IsContainPtr(cea) ) {
			if ( pvtable < Dword(cea) && min > Dword(cea) )
				min = Dword(cea);
			if ( pvtable == Dword(cea) )
				pvtable = 0;
			if ( cea >= ea + 0x2C )
				MakeSub(Dword(cea), "");
		}
	}

	pintf = ea + offs_INTF;					// interfaces table
	if ( IsContainPtr(pintf) )
		if ( (pintf = MakeINTFs(Dword(pintf))) > max )
			max = pintf;

	pxxx = ea + offs_XXX;
	if ( Dword(pxxx) != 0 ) {				// unknown field
		Warning("Unknown field in rtclass at 0x%08lX\n", ea);
		return BADADDR;
	}

	pprops = ea + offs_PROPS;				// properties table
	if ( IsContainPtr(pprops) )
		if ( (pprops = MakePROPs(Dword(pprops))) > max )
			max = pprops;

	pacls = ea + offs_ACLS;					// accelerators table
	if ( IsContainPtr(pacls) )
		if ( (pacls = MakeACLs(Dword(pacls))) > max )
			max = pacls;

	pmnus = ea + offs_MNUS;					// menus table
	if ( IsContainPtr(pmnus) )
		if ( (pmnus = MakeMNUs(Dword(pmnus))) > max )
			max = pmnus;

	pmsgs = ea + offs_MSG;					// messages table
	if ( IsContainPtr(pmsgs) )
		if ( (pmsgs = MakeMSGs(Dword(pmsgs))) > max )
			max = pmsgs;

	if ( pvtable != 0 && IsContainPtr(pvtable) ) {	// vtable
		do {
			MakeSub(Dword(pvtable), "");
			pvtable = MyMakeOff(pvtable);
		} while ( pvtable < min );
	}
	if ( pvtable > max )
		max = pvtable;

	return max;
}

// -----------------------------------------
// 
static CheckRptCmt()
{
	auto ea, cmt, i, nea, ask;

	for ( ea = FirstSeg(); ea != BADADDR; ea = NextFunction(ea) ) {
		cmt = GetFunctionCmt(ea, 1);
		if ( cmt != "" ) {
			if ( (i = strstr(cmt, "doubtful name")) != -1 ) {
				if ( i == 0 )
					SetFunctionCmt(ea, "", 1);
				else {
					cmt = substr(cmt, 0, i - 1);
					if ( strlen(cmt) != 0 ) {
						if ( (nea = LocByName(cmt)) == BADADDR )
							MyMakeName(nea, cmt);
						else {
							Jump(ea);
							Message(form("This function name '' exist at 0x%08lX\n", cmt, nea));
							SetFunctionCmt(ea, form("%s\nsee at 0x%08lX",cmt, nea), 1);
						}
					}
				}
			} else {
				if ( (nea = LocByName(cmt)) == BADADDR )
					MyMakeName(nea, cmt);
				else {
					Jump(ea);
					ask = AskYN(-1, form("Delete function comment at 0x%08lX?\n'%s'", ea, cmt));
					if ( ask == -1 )
						return;
					if ( ask == 1 )
						SetFunctionCmt(ea, "", 1);
					else {
						Message(form("This function name '' exist at 0x%08lX\n", cmt, nea));
						SetFunctionCmt(ea, form("%s\nsee at 0x%08lX",cmt, nea), 1);
					}
				}
			}
		}
	}
	Message(" ---> Check repeatable comment complit!\n");
}

// -----------------------------------------
//
static CheckExcept(ea, end)
{
	if ( (Byte(end) == 0xE9 && Byte(end + 0x05) == 0xEB) ||
		 (Byte(end) == 0xE9 && Byte(end + 0x05) == 0xE9) ) {
		for ( ;; ) {
			MyMakeUnkn(end, 8);
			if ( !MakeCode(end) ) {
				Jump(end);
				Message(form(" --->ERROR: Make exception at 0x%08lX???\n", end));
				return end;
			}
			if ( GetMnem(end) == "retn" ) {
				end = ItemEnd(end);
				if ( !SetFunctionEnd(ea, end) ) {
					Jump(end);
					Message(form(" --->ERROR: SetFunctionEnd(0x%08lX, 0x%08lX)\n", ea, end));
					return end;
				}
				return ea;
			} else
				end = ItemEnd(end);
		}
	}
	return end;
}

// -----------------------------------------
//	make '.text' segment
static MakeTextSeg()
{
	auto ea, end, sea, ask, name;

	ea = (ScreenEA() + 0x03) & ~0x03;
	end = SegEnd(ea);

	while ( ea < end ) {
		if ( isTPDSC(ea) ) {				// is type descriptor
			if ( (sea = MakeTPDSC(ea)) == BADADDR ) {
				Jump(ea);
				Message(form(" ---> ERROR: MakeTPDSC at 0x%08lX\n", ea));
				break;
			}
			ea = sea;
		} else if ( isRTDSC(ea) ) {			// is runtime class descriptor
			if ( (sea = MakeRTDSC(ea)) == BADADDR ) {
				Jump(ea);
				Message(form(" ---> ERROR: MakeRTDSC at 0x%08lX\n", ea));
				break;
			}
			ea = sea;
		} else if ( isInt(ea) ) {			// is Int
			ea = MakeInt(ea, "");
		} else if ( isChar(ea) ) {			// is Char
			ea = MakeChar(ea, "");
		} else if ( isEnum(ea) ) {			// is Enum
			if ( (sea = MakeEnum(ea, "")) == BADADDR ) {
				Jump(ea);
				Message(form(" ---> ERROR: MakeEnum at 0x%08lX\n", ea));
				break;
			}
			ea = sea;
		} else if ( isFloat(ea) ) {			// is Float
			ea = MyMakeFloat(ea, "");
		} else if ( isString(ea) ) {		// is String
			ea = MyMakeString(ea, "");
		} else if ( isSet(ea) ) {			// is Set
			ea = MakeSet(ea, "");
		} else if ( isClass(ea) ) {			// is Class
			ea = MakeClass(ea, "");
		} else if ( isMethod(ea) ) {		// is Method
			ea = MakeMethod(ea, "");
		// Byte(ea) == 0x09
		} else if ( isLString(ea) ) {		// is long String
			ea = MakeLString(ea, "");
		} else if ( isWString(ea) ) {		// is Wide string
			ea = MakeWString(ea, "");
		} else if ( isVariant(ea) ) {		// is Variant
			ea = MakeVariant(ea, "");
		} else if ( isArray(ea) ) {			// is Array
			ea = MyMakeArray(ea, "");
		} else if ( isRecord(ea) ) {		// is Record
			ea = MakeRecord(ea, "");
		} else if ( isInterface(ea) ) {		// is Interface
			ea = MakeInterface(ea, "");
		// Byte() == 0x10
		} else if ( isDynArray(ea) ) {		// is BCC Dyn array
			ea = MakeDynArray(ea, "");
		} else if ( isAlign(ea) ) {			// is Align
			ea = MyMakeAlign(ea);
		} else if ( isBCCString(ea) ) {		// is BCC string
			ea = MakeBCCString(ea);
		} else if ( (name = GetFunctionName(ea)) != "" ) {
			if ( strstr(name, "unknown_libname_") != -1 || strstr(name, "nullsub_") != -1 ) {
				MakeName(ea, "");
			}
			ea = CheckExcept(ea, FindFuncEnd(ea));
		} else {
			if ( Byte(ea) == 0x55 && Byte(ea + 1) == 0x8B ) {
				if ( MakeSub(ea, "") )
					ea = CheckExcept(ea, FindFuncEnd(ea));
				else {
					Jump(ea);
					Message(form(" ---> ERROR: Unknown object type at 0x%08lX\n", ea));
					break;
				}
			} else {
				Jump(ea);
				ask = AskYN(1, form("Unknown data at 0x%08lX\nMake Function -Yes\nSkip this - No\nCancel - Cancel?", ea));
				if ( ask == 1 ) {
					if ( MakeSub(ea, "") )
						ea = CheckExcept(ea, FindFuncEnd(ea));
					else {
						Jump(ea);
						Message(form(" ---> ERROR: Unknown object type at 0x%08lX\n", ea));
						break;
					}
				} else if ( ask == -1 )
					break;
				else {
					ea = ItemEnd(ea);
				}
			}
		}
	}
}

// -----------------------------------------
//	make '.data' segment
static MakeDataSeg()
{
	auto start, ea, end, iea;

	start = SegStart(ScreenEA());
	Message(form(" ---> Start at 0x%08lX\n", start));
	for ( ea = start; Dword(ea) != start; ea = ea + 4 ) {
		;
	}
	MyMakeName(Dword(ea), "__Initialize");
	end = Dword(ea + 4);
	for ( iea = Dword(ea); iea < end; ) {
		iea = MyMakeWord(iea);
		MakeSub(Dword(iea), "");
		iea = MyMakeOff(iea);
	}
	ea = MyMakeOff(ea);
	MyMakeName(Dword(ea), "__Finalize");
	end = Dword(ea + 8);
	for ( iea = Dword(ea); iea < end; ) {
		iea = MyMakeWord(iea);
		MakeSub(Dword(iea), "");
		iea = MyMakeOff(iea);
	}
	ea = MyMakeOff(ea);		// "__Finalize"
	ea = MyMakeOff(ea);		// "__Finalize"
	ea = MyMakeOff(ea);		
	ea = MyMakeDword(ea);
	ea = MyMakeDword(ea);
//	MyMakeName(Dword(ea), "WinMain");	// ??? Dll
	ea = MyMakeOff(ea);
//	MyMakeName(Dword(ea), "__matherr");
	ea = MyMakeOff(ea);
//	MyMakeName(Dword(ea), "__matherr1");
	ea = MyMakeOff(ea);
	Jump(ea);
	Message(" ---> Data\n");
}

// -----------------------------------------
//
static main()
{
	auto ea, name;

	BCCEnums();
	BCCStruct();

	ea = ScreenEA();
	name = SegName(ea);

	if ( name == ".text" )
		MakeTextSeg();
	else if ( name == ".data" )
		MakeDataSeg();
}
