//
//	Written by Dmitry Friesen, 20/03/2003
//	e-mail: dfriesen1946@mail.ru
//

//------------------------------------------------------------------------
//	IDA miscellaneous functions
//

#ifndef _IDC_IDC_
#include "idc.idc"
#define _IDC_IDC_
#endif	// _IDC_IDC_

#ifndef _IDA_MISC_IDC_
#define _IDA_MISC_IDC_

// -----------------------------------------
// if yes, the byte is defined.
// if no,  the byte is uninitialized.
static IsDefined(ea)
{
	return 0 != ( FF_IVL & GetFlags(ea) );
}

// -----------------------------------------
// if yes, the bytes in area defined.
// if no,  the any byte in area undefined.
static IsDefinedArea(ea, size)
{
	for ( ; size > 0; size-- )
		if ( !IsDefined(ea++) )
			return 0;

	return 1;
}

// -----------------------------------------
// if yes, 'ea' contain pointer to defined byte
// 
static IsContainPtr(ea)
{
	if ( !IsDefinedArea(ea, 4) || SegStart(Dword(ea)) == BADADDR )
		return 0;

	return IsDefined(Dword(ea));
}

// -----------------------------------------
// Make unknown area(ea..ea+size-1):
static MyMakeUnkn(ea, size)
{
	for ( ; size > 0; size-- ) {
		MakeUnkn(ea, 0);
		ea++;
	}
}

// -----------------------------------------
//
static MyMakeByte(ea)
{
	MyMakeUnkn(ea, 1);
	MakeByte(ea);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MyMakeWord(ea)
{
	MyMakeUnkn(ea, 2);
	MakeWord(ea);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MyMakeDword(ea)
{
	MyMakeUnkn(ea, 4);
	MakeDword(ea);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MyMakeOff(ea)
{
	MyMakeUnkn(ea, 4);
	OpOff(ea, 0, 0x0);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MyMakeDArray(ea, cnt)
{
	MyMakeUnkn(ea, 4 * cnt);
	MakeDword(ea);
	MakeArray(ea, cnt);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MyMakeSArray(ea, cnt, size)
{
	MyMakeUnkn(ea + size, size * (cnt-1));
	MakeArray(ea, cnt);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static isAlpha(c)
{
	return ( ( c >= 'a' && c <= 'z' ) || ( c >= 'A' && c <= 'Z' ) );
}

// -----------------------------------------
//
static isDigit(c)
{
	return ( c >= '0' && c <= '9' );
}

// -----------------------------------------
//
static isDelim(c)
{
	return ( c == '_' || c == ':' || c == ',' || c == '*' || c == '&' || \
			 c == '.' || c == '?' || c == '@' || c == '$' || \
			 c == '(' || c == ')' || \
			 c == '[' || c == ']' || \
			 c == '<' || c == '>' );
}

// -----------------------------------------
//
static isCStrChar(c)
{
	return ( isAlpha(c) || isDigit(c) || isDelim(c) );
}

// -----------------------------------------
//
static isCStr(ea)
{
	auto len, chr;

	chr = Byte(ea);
	for ( len = 0; isCStrChar(chr); len++ ) {
		ea++;
		chr = Byte(ea);
	}

	return ((chr == ' ' || chr == 0 ) && len >= 1);
}

// -----------------------------------------
//
static GetCStr(ea)
{
	auto name;

	name = "";
	while ( isCStrChar(Byte(ea)) ) {
		name = name + Byte(ea);
		ea++;
	}
	return name;
}

// -----------------------------------------
//
static MyMakeName(ea, name)
{
	auto adr, ask;

	if ( (adr = LocByName(name)) != BADADDR && adr != ea ) {
		Jump(adr);
		ask = AskYN(-1, form("Delete existing name?\n%s\nNew address is 0x%08lX", name, ea));
		if ( ask == 1 )
			MakeName(adr, "");
		else
			return 0;
	}

	if ( strstr(Name(ea), "?") != 0 )
		MakeName(ea, name);

	return 1;
}

// -----------------------------------------
//
static CStrLen(ea)
{
	auto len, chr;

	chr = Byte(ea);
	for ( len = 0; isCStrChar(chr) || chr == ' '; len++ ) {
		ea++;
		chr = Byte(ea);
	}
	return len;
}

// -----------------------------------------
//
static MakeCStr(ea, name)
{
	auto type;

	MyMakeUnkn(ea, CStrLen(ea) + 1);
	if ( (GetFlags(ea) & DT_TYPE) != FF_ASCI ) {
		type = GetLongPrm(INF_STRTYPE);
		SetLongPrm(INF_STRTYPE, ASCSTR_TERMCHR);
		MakeStr(ea, BADADDR);
		SetLongPrm(INF_STRTYPE, type);
	}
	MyMakeName(ea, name);
}

// -----------------------------------------
//
static IsPasStr(ea)
{
	auto n, ch;

	n = Byte(ea);
	for ( ; n > 0; n-- ) {
		ch = Byte(++ea);
		if ( !(ch == '_' || isAlpha(ch) || isDigit(ch)) )
			return 0;
	}
	return 1;
}

// -----------------------------------------
//
static MakePasStr(ea)
{
	auto type, len, name;

	type = GetLongPrm(INF_STRTYPE);
	SetLongPrm(INF_STRTYPE, ASCSTR_PASCAL);
	len = Byte(ea);
	MyMakeUnkn(ea, len + 1);
	if ( MakeStr(ea, BADADDR) == 0 ) {
		SetLongPrm(INF_STRTYPE, type);
		return "";
	}
	SetLongPrm(INF_STRTYPE, type);
	MyMakeName(ea, "");
	name = "";
	for ( ; len > 0; len-- )
		name = name + Byte(++ea);

	return name;
}

// -----------------------------------------
//
static isAlign(ea)
{
	auto bt;

	if ( (ea & 0x3) != 0) {
		bt = Byte(ea);
		if ( bt == 0x00 )
			return 1;
		else if ( bt == 0x90 )
			return 1;
		else if ( bt == 0xCC )
			return 1;
		else if ( bt == 0x8B && Byte(ea+1) == 0xC0 )
			return 1;
		else if ( bt == 0x8D && Byte(ea+1) == 0x40 && Byte(ea+2) == 0x00 )
			return 1;
	}
	return 0;
}

// -----------------------------------------
//
static MyMakeAlign(ea)
{
	auto start, bt, n;

	start = ea;
	bt = Byte(ea);
	if ( bt == 0x00 ) {
		for ( n = 0; (ea & 0x03) != 0 && Byte(ea) == 0x00; n++ )
			ea++;
		MyMakeUnkn(start, n);
		MakeAlign(start, n, 0);
	} else if ( bt == 0x90 ) {
		for ( n = 0; (ea & 0x03) != 0 && Byte(ea) == 0x90; n++ )
			ea++;
		MyMakeUnkn(start, n);
		MakeAlign(start, n, 0);
	} else if ( bt == 0xCC ) {
		for ( n = 0; (ea & 0x03) != 0 && Byte(ea) == 0xCC; n++ )
			ea++;
		MyMakeUnkn(start, n);
		MakeAlign(start, n, 0);
	} else if ( bt == 0x8B && Byte(ea + 1) == 0xC0 ) {
		MyMakeUnkn(start, 2);
		MakeAlign(start, 2, 0);
		n = 2;
	} else if ( bt == 0x8D && Byte(ea + 1) == 0x40 && Byte(ea + 2) == 0x00 ) {
		MyMakeUnkn(start, 3);
		MakeAlign(start, 3, 0);
		n = 3;
	}

	return start + n;
}

// -----------------------------------------
// Make structure by name:
static MyMakeStruct(ea, name)
{
	auto sid, size;

	sid = GetStrucIdByName(name);
	size = GetStrucSize(sid);
	MyMakeUnkn(ea, size);
	MakeStruct(ea, name);
	return ItemEnd(ea);
}

// -----------------------------------------
//
static MySetFuncCmt(ea, name, rep)
{
	auto cmt;

	cmt = GetFunctionCmt(ea, rep);
	if ( strstr(cmt, name) == -1 ) {
		if ( cmt != "" )
			SetFunctionCmt(ea, name, rep);
		else
			SetFunctionCmt(ea, cmt + "\n" + name, rep);
	}
}

// -----------------------------------------
//
static MyMakeFunction(ea)
{
	auto start, end, n, uea, ask;

	start = ea;
	end = NextFunction(start);
	if ( end == -1 )
		end = SegEnd(start);

	if ( !isCode(GetFlags(ea)) ) {
		MakeUnkn(ea, 1);
		MakeCode(ea);
	}
	while ( MakeFunction(ea, BADADDR) == 0 ) {
		while ( start != end ) {
			if ( Byte(start) == 0x90 ) {
				if ( isCode(GetFlags(start + 1)) ) {
					Jump(start);
					ask = AskYN(1, form("This NOP is in function or NOT?", start));
					if ( ask != 1 )
						return ( MakeFunction(ea, BADADDR) != 0 );
				} else
					return ( MakeFunction(ea, BADADDR) != 0 );
			}
			if ( isCode(GetFlags(start)) )
				start = ItemEnd(start);
			else {
				uea = start;
				while ( (n = MakeCode(start)) == 0 ) {
					if ( uea == end )
						return ( MakeFunction(ea, BADADDR) != 0 );
					MakeUnkn(uea, 0);
					uea++;
				}
				start = start + n;
				break;
			}
		}
	}

	return 1;
}

// -----------------------------------------
//
static MakeSub(ea, name)
{
	auto sub, end;

	if ( (GetFlags(ea) & FF_FUNC) != FF_FUNC )
		if ( !MyMakeFunction(ea) )
			return 0;

	sub = Name(ea);
	if ( strstr(sub, "sub_") == 0 ) {
		if ( name != "" )
			MyMakeName(ea, name);
	} else {
		if ( GetFunctionCmt(ea, 0) == "" ) {
			if ( name != "" )
				MyMakeName(ea, name);
		}
	}

	return 1;
}

#endif	// _IDA_MISC_IDC_
